# input.s: Various input handling routines are in this file.
# The routines handle input from two sources (serial and keypad)
# and put both into an input buffer stored in RAM.
# This buffer may be read by high level routines.

# Global procedures and functions defined in this file:
.globl keypad_input_received
.globl serial_input_received
.globl init_input_buffer
.globl receive_character
.globl toggle_input

# Global procedures and functions called in this file:

# Constants and definitions about the key buffer:
key_buffer_length_bits = 4
key_buffer_length = 16          # = 2^length_bits.



.bss

# This is the space for the key buffer

key_buffer:     
                .space      key_buffer_length

# Here are the two pointers for reading and
# writing to the buffer.

key_buffer_read_ptr:        
                .space      1
key_buffer_write_ptr:        
                .space      1

# This flag is set if and only if input should be
# accepted. It may be changed by toggle_input.
input_switched_on:        
                .space      1

.text

# Write data into the buffer:
#   - IF key_buffer_write_ptr = key_buffer_read_ptr THEN full.
#   - write new character to (key_buffer + key_buffer_write_ptr)
#   - key_buffer_write_ptr ++ ;
#
# Read data from buffer
#   - IF key_buffer_write_ptr = key_buffer_read_ptr + 1 THEN empty.
#   - key_buffer_read_ptr ++ ;
#   - read character from (key_buffer + key_buffer_read_ptr)

# keypad_input_received: Input handler.
# serial_input_received: Input handler.
# These procedures are called whenever a button is pressed on
# the keypad or when a character is received on the serial line.
# Both do exactly the same thing and take the same input,
# so the two are equivalent.
# 
# They must ONLY be called when interrupts are disabled to avoid
# concurrency issues with buffer access.

keypad_input_received:
serial_input_received:
                push        ix
                ld          ix,0x0000
                add         ix,sp

                push        af
                push        bc
                push        hl

                # Is input enabled?
                # Load flag into A...
                ld          a,(input_switched_on)
                cp          0       # is it zero?
                jr          z,input_is_disabled

                # Is the buffer full?
                ld          a,(key_buffer_write_ptr)
                ld          c,a             # put write ptr in C

                ld          a,(key_buffer_read_ptr)

                # If the buffer is full, we have to discard this
                # character :(
                cp          c
                jr          z,buffer_full

                # Find the write position: key_buffer + key_buffer_write_ptr
                # Do this by loading key_buffer into HL, and adding BC to it.
                ld          hl,key_buffer
                ld          b,0
                add         hl,bc

                ld          a,(ix+5)        # load character pressed
                ld          (hl),a          # put character in buffer

                ld          a,c             # store write ptr in A
                # increment write pointer with modulo key_buffer_length
                inc         a               
                and         key_buffer_length - 1

                # store write ptr.
                ld          (key_buffer_write_ptr),a

input_is_disabled:
buffer_full:
                # done...
                pop         hl
                pop         bc
                pop         af
                pop         ix
                ret


# receive_character: Function to get a character from the input buffer,
# if there is one there. If there are no characters in the input
# buffer, this function returns zero.
# Usage:
#     push af
#     call receive_character
#     pop af
# a contains the character read or zero.
receive_character:
                push        ix
                ld          ix,0x0000
                add         ix,sp

                push        af
                push        bc
                push        hl

                # Return zero by default
                ld          (ix+5),0

                # Disable interrupts before accessing the buffer
                # code to prevent concurrency issues!
                di

                # store the write pointer in L.
                ld          a,(key_buffer_write_ptr)
                ld          l,a

                # Is the buffer empty?
                ld          a,(key_buffer_read_ptr)
                # increment read pointer with modulo key_buffer_length
                inc         a               
                and         key_buffer_length - 1

                # if the new read pointer = the write pointer,
                # then the buffer is empty. The write pointer
                # was stored in L earlier.
                cp          l
                jr          z,buffer_empty

                # if not, store new read_ptr.
                ld          (key_buffer_read_ptr),a

                # and load character from buffer
                ld          hl,key_buffer
                ld          b,0
                ld          c,a
                add         hl,bc

                ld          a,(hl)
                ld          (ix+5),a

buffer_empty:   # re-enable interrupts
                ei

                pop         hl
                pop         bc
                pop         af
                pop         ix
                ret


# init_input_buffer: Initialisation procedure.
# Sets up the ring buffer variables, and switch on the
# input. This procedure takes no parameters.
init_input_buffer:
                push        af

                ld          a,key_buffer_length - 1
                ld          (key_buffer_read_ptr),a
                ld          a,0
                ld          (key_buffer_write_ptr),a

                inc         a       # set A = 1
                ld          (input_switched_on),a

                pop         af

                ret

# toggle_input: Use this procedure to tell the input
# subsystem whether it should accept input or not. It
# should be called with a single parameter: if that
# parameter is non-zero, then input is sampled.
#
# Usage: ld a,<accept-input>
#        push af
#        call toggle_input
#        pop af
toggle_input:   di          # disable interrupts
                push        ix
                ld          ix,0x0000
                add         ix,sp

                push        af
                ld          a,(ix+5)        # load in input flag.
                ld          (input_switched_on),a   # set global flag.
                pop         af

                pop         ix
                ei                          # re-enable interrupts
                ret


